'''
Bind to an LDAP directory and perform various operations.
'''

import sys
import getpass
import copy
import re
import ipaddress

import ldap3
from ldap3.core.exceptions import LDAPSocketOpenError, LDAPStartTLSError, \
    LDAPSessionTerminatedByServerError, LDAPSocketReceiveError

from .ldif import ldif_read
from .password import ssha_passwd
from .config import config
from .terminal import fmt


def ping(uri):
    '''
    Returns true if an LDAP server is responding at a given URI by attempting
    an anonymous bind.
    '''
    try:
        uri = clean_uri(uri)
        con = ldap3.Connection(uri, auto_bind=ldap3.AUTO_BIND_TLS_BEFORE_BIND)
        con.unbind()
        return True
    except (LDAPSocketOpenError, LDAPSessionTerminatedByServerError,
        LDAPSocketReceiveError):
        return False


def supports_starttls(uri):
    '''
    Determine if the server actually supports StartTLS (both the server software
    itself supports it, and the server instance itself has been configured with
    SSL support).
    '''
    try:
        uri = clean_uri(uri)
        con = ldap3.Connection(uri, auto_bind=ldap3.AUTO_BIND_TLS_BEFORE_BIND)
        con.unbind()
        return True
    except LDAPStartTLSError:
        return False


def auto_bind(conf=None, server_info=True):
    '''
    Automatically detects LDAP config values and returns a directory binding.
    '''
    if conf is None:
        conf = config()

    if conf['binddn'] is not None and conf['bindpw'] is None:
        conf['bindpw'] = getpass.getpass('Enter bind DN password...')

    return Connection(conf['host'], user=conf['binddn'],
        password=conf['bindpw'], conf=conf)


def dn_address(dn):
    '''
    Get the "."-delmited address for a DN (typically a directory naming context/
    base dn). If the directory naming context was dc=ezldap,dc=io, then the
    address would be "ezldap.io". Typically used when generating fully-qualified
    hostnames for new hosts (for instance, "hostname.ezldap.io"). However,
    this function will create addresses for any DN, if so desired.
    '''
    dn = re.sub(r'\s', '-', dn.lower().strip())
    contents = re.findall(r'=([^=,]+)', dn)
    return '.'.join(contents)


def clean_uri(uri):
    '''
    ldap3 really struggles with URIs ending in a slash.
    '''
    uri = re.sub(r'///', '//localhost', uri.strip())
    return re.sub(r'/$', '', uri)


class Connection(ldap3.Connection):
    '''
    An object-oriented wrapper around an LDAP connection.
    Used to make pyldap's LDAPObject even easier to use.
    To automatically create a binding use auto_bind() instead.
    '''

    def __init__(self, host, user=None, password=None, conf=None,
        authentication=ldap3.SIMPLE, server_info=True):
        '''
        Create a new LDAP connection and bind. If either user or password are
        omitted, the bind is anonymous. conf is a dictionary with any
        placeholders or key/value combinations that you wish to be passed to
        ldif templates or details like user/group OUs.
        '''
        if host is None:
            raise ValueError('LDAP host cannot be None.')

        if conf is None:
            conf = config()

        # delete bind credentials from config, don't want those after binding
        conf.pop('binddn', None)
        conf.pop('bindpw', None)
        self.conf = conf

        # for whatever reason, ldap3 can't deal with ldap:/// identifiers
        host = clean_uri(host)
        if server_info:
            self.server = ldap3.Server(host, get_info=ldap3.ALL)
        else:
            self.server = ldap3.Server(host, get_info=ldap3.NONE)

        if supports_starttls(host):
            auto_bind_mode = ldap3.AUTO_BIND_TLS_BEFORE_BIND
        else:
            print(fmt('Warning: server does not appear to support SSL/StartTLS, '
                'proceeding without...', color='yellow'), file=sys.stderr)
            auto_bind_mode = ldap3.AUTO_BIND_NO_TLS

        if user is None or password is None:
            # anonymous bind
            super().__init__(self.server, auto_bind=auto_bind_mode)
        else:
            super().__init__(self.server,
                user=user, password=password, authentication=authentication,
                auto_bind=auto_bind_mode)


    def __enter__(self):
        return self


    def __exit__(self, type_, value, traceback):
        '''
        Auto-unbind when used with "with" keyword.
        '''
        self.unbind()


    def who_am_i(self):
        '''
        Easier wrapper around who_am_i().
        '''
        return self.extend.standard.who_am_i()


    def base_dn(self):
        """
        Detect the base DN from an LDAP connection.
        """
        return self.server.info.naming_contexts[0]


    def search_list(self, search_filter='(objectClass=*)',
                    attributes=ldap3.ALL_ATTRIBUTES, search_base=None, **kwargs):
        '''
        A wrapper around search() with better defaults and output format.
        A list of dictionaries will be returned, with one dict per output object.
        If search_base is None, the directory base DN will be used.
        '''
        if search_base is None:
            search_base = self.base_dn()

        self.search(search_base, search_filter, attributes=attributes, **kwargs)
        query = []
        for res in self.response:
            result = {'dn': [res['dn']]}
            # ensure every attribute is encapsulated in a list
            for k, v in res['attributes'].items():
                if not isinstance(v, list):
                    res['attributes'][k] = [v]

            result.update(res['attributes'])
            query.append(result)

        return query


    def search_list_t(self, search_filter='(objectClass=*)',
                      attributes=ldap3.ALL_ATTRIBUTES, search_base=None,
                      unpack_lists=True, unpack_delimiter='|', **kwargs):
        '''
        A utility function that returns the transposed result of search_list()
        (a dict of lists, with one list per attribute.)
        This is very useful for tasks like retrieving all uidNumbers currently
        assigned or emails used by users. The DN of each entry is always output.
        '''
        response = self.search_list(search_filter, attributes=attributes,
            search_base=search_base, **kwargs)

        all_attribs = {'dn'}
        if attributes == ldap3.ALL_ATTRIBUTES:
            # iterate through and determine all possible attributes returned
            for res in response:
                all_attribs.update(res.keys())
        elif attributes is None:
            # skip over this case, we do not want a list of None
            pass
        else:
            if not isinstance(attributes, list):
                attributes = [attributes]

            all_attribs.update(attributes)

        query = {attrib: [] for attrib in all_attribs}

        for res in response:
            for k in all_attribs:
                try:
                    v = res[k]
                    if unpack_lists and isinstance(v, list):
                        if len(v) > 1:
                            # coerces to string
                            v = unpack_delimiter.join([str(x) for x in v])
                        else:
                            # string coercion avoided for 1-element lists
                            v = v[0]

                except KeyError:
                    v = None

                query[k].append(v)

        return query


    def search_df(self, search_filter='(objectClass=*)',
                  attributes=ldap3.ALL_ATTRIBUTES, search_base=None, **kwargs):
        '''
        A convenience function to search an LDAP directory and return a Pandas
        DataFrame. Very useful for analyzing the contents of your directory,
        computing stats, etc. Requires the pandas package to be installed.
        '''
        try:
            from pandas import DataFrame
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError('This function requires the pandas package to be installed.') from e

        query = self.search_list_t(search_filter, attributes=attributes,
            search_base=search_base, **kwargs)
        return DataFrame(query)


    def exists(self, dn):
        '''
        Returns true if a given DN exists in an LDAP directory.
        '''
        query = self.search_list(search_base=dn, search_scope=ldap3.BASE)
        if len(query) == 1:
            return True
        else:
            return False


    def next_uidn(self, search_filter='(objectClass=posixAccount)',
        search_base=None, uid_start=10000, uid_attribute='uidNumber'):
        """
        Determine the next available uid number in a directory tree.
        """
        users = self.search_list_t(search_filter, uid_attribute, search_base=search_base)
        if len(users['dn']) == 0:
            return uid_start

        return max(users[uid_attribute]) + 1


    def next_gidn(self, search_filter='(objectClass=posixGroup)',
        search_base=None, gid_start=10000, gid_attribute='gidNumber'):
        """
        Determine the next available gid number in a directory tree.
        """
        return self.next_uidn(search_filter, search_base, gid_start, gid_attribute)


    def _conf_basedn_key(self, key):
        '''
        A helper to fetch basedn values for several search functions.
        '''
        try:
            return self.conf[key]
        except KeyError:
            return self.base_dn()


    def get_user(self, user, basedn=None, index='uid'):
        '''
        Return given user as a dict or None if none is found. Searches entire
        directory if no base search dn given.
        '''
        if basedn is None:
            basedn = self._conf_basedn_key('peopledn')

        try:
            return self.search_list('({}={})'.format(index, user), search_base=basedn)[0]
        except IndexError:
            return None


    def get_group(self, group, basedn=None, index='cn'):
        '''
        Return a given group. Searches entire directory if no base search dn given.
        '''
        if basedn is None:
            basedn = self._conf_basedn_key('groupdn')

        return self.get_user(group, basedn=basedn, index=index)


    def get_host(self, host, basedn=None, index='cn'):
        '''
        Return a given host. Searches entire directory if no base search dn given.
        '''
        if basedn is None:
            basedn = self._conf_basedn_key('hostsdn')

        return self.get_user(host, basedn=basedn, index=index)


    def ldif_add(self, ldif):
        """
        Perform an add operation using an LDIF object.
        """
        results = []
        for entry in ldif:
            entry_cp = copy.deepcopy(entry)
            dn = entry_cp.pop('dn')[0]
            #TODO fix for 389 directory server and "objectClasses"
            objectclass = entry_cp.pop('objectClass')
            self.add(dn=dn, object_class=objectclass, attributes=entry_cp)
            results.append(self.result)

        return results


    def ldif_modify(self, ldif):
        """
        Perform an LDIF modify operation from an LDIF object.
        """
        results = []
        for entry in ldif:
            entry_cp = copy.deepcopy(entry)
            dn = entry_cp.pop('dn')[0]
            self.modify(dn, entry_cp)
            results.append(self.result)

        return results


    def modify_replace(self, dn, attrib, value, replace_with=None):
        '''
        Change a single attribute on an object.
        '''
        if value is None:
            raise ValueError('value cannot be None when performing a replace operation.')

        if replace_with is None:
            self.modify(dn, {attrib: [(ldap3.MODIFY_REPLACE, [value] )] })
        else:
            # Delete then add seems to be the way to replace a specific value
            # using ldap3.
            self.modify_delete(dn, attrib, value)
            self.modify_add(dn, attrib, replace_with)

        return self.result


    def modify_add(self, dn, attrib, value):
        '''
        Add a single attribute to an object.
        '''
        if value is None:
            raise ValueError('value cannot be None when performing an add operation.')

        self.modify(dn, {attrib: [(ldap3.MODIFY_ADD, [value] )] })
        return self.result


    def modify_delete(self, dn, attrib, value=None):
        '''
        Delete a single attribute from an object.
        If value is None, deletes all attributes of that name.
        '''
        if value is None:
            self.modify(dn, {attrib: [(ldap3.MODIFY_DELETE, [] )] })
        else:
            self.modify(dn, {attrib: [(ldap3.MODIFY_DELETE, [value] )] })

        return self.result


    def add_group(self, groupname,
        ldif_path='~/.ezldap/add_group.ldif', **kwargs):
        """
        Adds a group from an LDIF template.
        """
        replace = {'groupname': groupname, 'gid': self.next_gidn()}
        replace.update(self.conf)
        replace.update(kwargs)
        ldif = ldif_read(ldif_path, replace)
        return self.ldif_add(ldif)


    def add_to_group(self, username, groupname,
        ldif_path='~/.ezldap/add_to_group.ldif', **kwargs):
        """
        Adds a user to a group.
        The user and group in question must already exist.
        """
        replace = {'username': username, 'groupname': groupname}
        replace.update(self.conf)
        replace.update(kwargs)
        ldif = ldif_read(ldif_path, replace)
        return self.ldif_modify(ldif)


    def add_user(self, username, groupname, password,
        ldif_path='~/.ezldap/add_user.ldif', **kwargs):
        '''
        Adds a user. Does not create or modify groups.
        "groupname" may be None if "gid" is specified.
        '''
        replace = {'username': username,
                   'user_password': ssha_passwd(password),
                   'uid': self.next_uidn(),
                   'gid': None}

        replace.update(self.conf)
        replace.update(kwargs)
        if replace['gid'] is None:
            try:
                replace['gid'] = self.get_group(groupname)['gidNumber'][0]
            except IndexError:
                raise ValueError('Group does not exist')

        ldif = ldif_read(ldif_path, replace)
        return self.ldif_add(ldif)


    def add_host(self, hostname, ip_address,
        ldif_path='~/.ezldap/add_host.ldif', **kwargs):
        '''
        Add a host to a directory. Hostname is the short hostname (hostname -s),
        to add. If specifying the fully qualified hostname is desired
        (or the fully qualified hostname does not match the directory suffix),
        specify the fully-qualified hostname as "hostname_fq".
        '''
        replace = {'hostname': hostname,
                   'ip': str(ipaddress.ip_address(ip_address)),
                   'hostname_fq': hostname + '.' + dn_address(self.base_dn())}
        replace.update(self.conf)
        replace.update(kwargs)
        ldif = ldif_read(ldif_path, replace)
        return self.ldif_add(ldif)



"""
Tools for generating and hashing OpenLDAP-compatible passwords.
Using SHA1 since that is the OpenLDAP default.

https://gist.github.com/rca/7217540 used as reference.
"""

import os
import re
import string
import hashlib
import base64
from random import SystemRandom

def random_passwd(length=10, ambiguous_chars=False):
    """
    Generate a readable, random password with no ambiguous characters
    (unless you set that option to true, of course).
    """
    chars = string.ascii_letters + string.digits
    if not ambiguous_chars:
        chars = re.sub('[1lIO0]', '', chars)

    r = SystemRandom()
    return ''.join(r.choice(chars) for char in range(length))


def ssha(val, salt):
    """
    Generate an SSHA hash.
    """
    ecrypt = hashlib.sha1(val.encode())
    ecrypt.update(salt)
    return ecrypt


def ssha_passwd(str_val):
    """
    Hash and salt a string using SHA1 algorithm and format for use with LDAP.
    """
    salt = os.urandom(4)
    out = ssha(str_val, salt)
    out = '{SSHA}' + base64.b64encode(out.digest() + salt).decode()
    return out


def ssha_check(ssha_val, str_val):
    """
    Check that a password decodes to the correct password.
    """
    ssha_val = base64.b64decode(ssha_val[6:])
    # SHA1 hashes are 20 characters
    salt = ssha_val[20:]
    digest = ssha_val[:20]

    # create a hash from str_val using original salt and see if it matches
    str_val_ssha = ssha(str_val, salt)
    return digest == str_val_ssha.digest()



'''
Autodetect and retrieve configuration values from an active LDAP server.
Ideally the user should have to do nothing aside from specify a password.
'''

import os
import yaml

def config(path=None):
    '''
    Attempts to generate a dictionary of config values for LDAP details from
    the following config files, in order: the environment variable EZLDAP_CONFIG,
    ~/.ezldap/config.yml, or guess from /etc/openldap/ldap.conf + /usr/bin/ldapwhoami.
    '''
    if path is not None:
        return yaml.load(open(os.path.expanduser(path)))
    elif 'EZLDAP_CONFIG' in os.environ.keys():
        # EZLDAP_CONFIG will have already been expanded by the user's shell
        return yaml.load(open(os.environ['EZLDAP_CONFIG']))
    elif os.path.exists(os.path.expanduser('~/.ezldap/config.yml')):
        return yaml.load(open(os.path.expanduser('~/.ezldap/config.yml')))
    else:
        return guess_config()


def guess_config():
    base = get_ldap_conf_val('BASE')
    conf = {
        'host': get_ldap_conf_val('URI'),
        'binddn': 'cn=Manager,{}'.format(base),
        'bindpw': None,
        'peopledn': 'ou=People,{}'.format(base),
        'groupdn': 'ou=Group,{}'.format(base),
        'hostsdn': 'ou=Hosts,{}'.format(base),
        'homedir': '/home'
    }
    return conf


def readlines_to_dict(list_of_strings):
    '''
    Read a UNIX conf file or LDIF output as a key-value dictionary.
    '''
    conf = {}
    for line in list_of_strings:
        if line[0] in ['#', '\n']:
            continue
        else:
            line_l = line.split()
            key = line_l.pop(0)
            conf[key] = line_l

    return conf


def get_ldap_conf_val(field):
    if os.path.exists('/etc/openldap/ldap.conf'):
        # redhat distros
        path = '/etc/openldap/ldap.conf'
    elif os.path.exists('/etc/ldap/ldap.conf'):
        # ubuntu
        path = '/etc/ldap/ldap.conf'
    else:
        return None

    try:
        ldap_conf = readlines_to_dict(open(path).readlines())
        return ldap_conf[field][0]
    except KeyError:
        return None



'''
Utility functions for reading/writing to LDIF files.
'''

import os
import re
import copy
from io import StringIO
from string import Template

import ldap3


class LDIFTemplateError(KeyError):
    '''
    A custom exception to better catch LDIF template errors specifcally instead
    of the ultra-vague KeyError.
    '''
    pass


def template(path, replacements=None):
    '''
    Read a file and substitute replacment entries for placeholders designated
    by $placeholder_name. If replacements is None, it simply opens and reads a
    file into a string.
    '''
    if replacements is None:
        return open(path).read()
    else:
        try:
            content = Template(open(path).read())
            return content.substitute(replacements)
        except KeyError as e:
            raise LDIFTemplateError('No value provided for LDIF key "{}"'
                .format(e.args[0])) from None


def ldif_read(path, replacements=None):
    '''
    Read an LDIF file into a list of dicts appropriate for use with ezldap.
    '''
    # read into a string buffer first
    path = os.path.expanduser(path)
    content = StringIO(template(path, replacements))

    operations = {
        'add': ldap3.MODIFY_ADD,
        'replace': ldap3.MODIFY_REPLACE,
        'delete': ldap3.MODIFY_DELETE
    }

    entries = []
    entry = {}
    changetype = 'add'
    next_change_attr = None
    next_change_type = 'add'
    for line in content:
        if line[0] == '#':
            continue
        if line[0] == '-':
            if next_change_type == 'delete' and len(entry[next_change_attr]) == 0:
                # The last entry ended with no entries and was also a "delete"
                # change. This should result in a delete change that deletes all
                # entries for that attribute.
                entry[next_change_attr].append((ldap3.MODIFY_DELETE, []))

            continue
        elif re.match(r'dn:', line):
            # new dn- add last entry, and start a new one
            if 'dn' in entry.keys():
                entries.append(entry)
                changetype = 'add'

            entry = {}

        match = re.findall(r'(\w+):\s*(.+)', line)
        if len(match) > 0:
            key = match[0][0]
            value = match[0][1].strip()

            if key == 'changetype':
                # determine changetype and skip line
                changetype = value
                continue
            elif key not in entry.keys() and key not in operations.keys():
                entry[key] = []

            if changetype == 'modify':
                # handle ldif-change LDIFs
                if key in operations.keys():
                    next_change_type, next_change_attr = key, value
                    if value not in entry.keys():
                        # a check in case the attribute has not already been added
                        entry[value] = []

                    continue
                elif key == next_change_attr:
                    # each change is handled separately,
                    # not the most efficient, but is an easier implementation
                    value = (operations[next_change_type], [value])
                else:
                    raise ValueError('Attribute does not match attribute to {}.'.format(next_change_type))

            entry[key].append(value)

    # last ldif object won't be added otherwise
    if 'dn' in entry.keys():
        entries.append(entry)

    return entries


def ldif_write(entries, path):
    '''
    Write self.entries as LDIF file.
    '''
    with open(os.path.expanduser(path), 'w') as handle:
        _entries_to_handle(entries, handle)


def ldif_print(entries):
    '''
    Print an LDIF entry to stdout.
    '''
    with StringIO() as strbuf:
        _entries_to_handle(entries, strbuf)
        strbuf.seek(0)
        print(''.join(strbuf.readlines()), end='')


def _entries_to_handle(entries, handle):
    '''
    Write entries to a filehandle.
    '''
    # because we're modifying the entries contained as we iterate through.
    entries_cp = copy.deepcopy(entries)
    for entry in entries_cp:
        handle.writelines(_dump_attributes('dn', entry.pop('dn')))
        #TODO only works with ldif-add, needs the ability to handle ldif-change
        handle.writelines(_dump_attributes('objectClass', entry.pop('objectClass')))
        for k, v in entry.items():
            handle.writelines(_dump_attributes(k, v))

        handle.write('\n')


def _dump_attributes(key, values):
    '''
    Convert a dictionary key/value pair (key: [value1, value2]) to a list of the
    form: ['key: value1', 'key: value2'].
    '''
    if not isinstance(values, list):
        values = [values]

    out = []
    for v in values:
        out.append('{}: {}\n'.format(key, v))

    return out



from setuptools import setup

# import version from version.py so it's only specified in one spot
exec(open('ezldap/version.py').read())

setup(
    name='ezldap',
    version=__version__,
    description='Scripts and Python bindings for easy LDAP operations.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/jstaf/ezldap',
    author='Jeff Stafford',
    author_email='jeff.stafford@queensu.ca',
    license='BSD3',
    packages=['ezldap'],
    scripts=['bin/ezldap'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'PyYAML',
        'ldap3'
    ],
    setup_requires=['pytest-runner'],
    tests_require=[
        'pandas',
        'pytest',
        'pytest-docker',
        'pytest-cov',
        'docker-compose'
    ],
    classifiers=[
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Topic :: System :: Systems Administration :: Authentication/Directory :: LDAP',
        'License :: OSI Approved :: BSD License',
        'Development Status :: 4 - Beta'
    ]
)



'''
Helper functions for nicer terminal formatting using escape sequences.
'''

ANSI_ESCAPE = {
    'white': '',
    'purple': '\033[95m',
    'blue': '\033[94m',
    'green': '\033[92m',
    'yellow': '\033[93m',
    'red': '\033[91m',
    'end': '\033[0m',
    'bold': '\033[1m',
    'underline': '\033[4m'
}


def fmt(text, color='white', bold=False, underline=False):
    if bold:
        text = ANSI_ESCAPE['bold'] + text

    if underline:
        text = ANSI_ESCAPE['underline'] + text

    return ANSI_ESCAPE[color] + text + ANSI_ESCAPE['end']


def _ansi_demo():
    print(fmt('purple', color='purple'))
    print(fmt('blue', color='blue'))
    print(fmt('green', color='green'))
    print(fmt('yellow', color='yellow'))
    print(fmt('red', color='red'))
    print(fmt('bold', bold=True))
    print(fmt('underline', underline=True))
